from gscraper.base.types import _TYPE, TypeHint, Timezone, CastError
from gscraper.base.types import get_type, is_bool_type, is_float_type, is_int_type
from gscraper.base.types import is_datetime_type, is_time_type, is_timestamp_type, is_date_type
from gscraper.base.types import is_str_type, is_list_type, is_tuple_type, is_set_type
from gscraper.base.types import RegexFormat, MatchFunction
from gscraper.base.types import DateFormat, DateNumeric, Timestamp

from gscraper.utils import notna, exists

from typing import Any, Callable, List, Literal, Optional, Set, Tuple, Union
from dateutil.parser import parse as dateparse
from pytz import timezone, BaseTzInfo, UnknownTimeZoneError
import datetime as dt
import re


###################################################################
########################### Numeric Type ##########################
###################################################################

def isfloat(__object) -> bool:
    try: return isinstance(float(__object), float)
    except CastError: return False


def cast_float(__object, default=0., clean=True, trunc: Optional[int]=None) -> float:
    try:
        __object = float(re.sub(r"[^\d.-]",'',str(__object))) if clean else float(__object)
        return round(__object, trunc) if isinstance(trunc, int) else __object
    except CastError: return default


def cast_float2(__object, default=0., clean=True, trunc=2) -> float:
    return cast_float(__object, default, clean, trunc)


def cast_int(__object, default=0, clean=True) -> int:
    try: return int(float(re.sub(r"[^\d.-]",'',str(__object)))) if clean else int(float(__object))
    except CastError: return default


def cast_int1(__object, default=1, clean=True) -> int:
    return cast_int(__object, default, clean)


def cast_numeric(__object, __type: TypeHint="auto", default=0, clean=True,
                trunc: Optional[int]=None, **kwargs) -> Union[float,int]:
    if __type == "auto": pass
    elif is_int_type(__type): return cast_int(__object, default=default, clean=clean)
    elif is_float_type(__type): return cast_float(__object, default=default, clean=clean, trunc=trunc)
    else: return default

    __object = cast_float(__object, default=default, clean=clean, trunc=trunc)
    if isinstance(__object, float):
        return __object if __object % 1. else int(__object)


###################################################################
########################## Datetime Type ##########################
###################################################################

TS_SECONDS = 10
TS_MILLISECONDS = 13


def get_timezone(tzinfo: Optional[Timezone]=None) -> BaseTzInfo:
    if not tzinfo: return None
    try: return timezone(tzinfo)
    except UnknownTimeZoneError: return


def set_timezone(__datetime: dt.datetime, tzinfo: Optional[Timezone]=None,
                astimezone: Optional[Timezone]=None, droptz=False) -> dt.datetime:
    tzinfo = get_timezone(tzinfo)
    if tzinfo:
        __datetime = __datetime.astimezone(tzinfo) if __datetime.tzinfo else tzinfo.localize(__datetime)
    if astimezone:
        __datetime = __datetime.astimezone(get_timezone(astimezone))
    return __datetime.replace(tzinfo=None) if droptz else __datetime


def get_ts_unit(__timestamp: Timestamp) -> str:
    __timestamp = cast_int(__timestamp, clean=False)
    if len(str(__timestamp)) == TS_SECONDS: return "s"
    elif len(str(__timestamp)) == TS_MILLISECONDS: return "ms"


def from_timestamp(__timestamp: Union[Timestamp,str], default=None,
                    tzinfo: Optional[Timezone]=None, astimezone: Optional[Timezone]=None) -> dt.datetime:
    try:
        if isinstance(__timestamp, str): __timestamp = cast_numeric(__timestamp, default=None, clean=False)
        __timestamp = __timestamp/1000 if get_ts_unit(__timestamp) == "ms" else __timestamp
        return set_timezone(dt.datetime.fromtimestamp(__timestamp), tzinfo, astimezone)
    except CastError: return default


def cast_datetime(__object: DateFormat, default=None, tzinfo: Optional[Timezone]=None,
                    astimezone: Optional[Timezone]=None, droptz=False) -> dt.datetime:
    try:
        if isinstance(__object, dt.datetime): __datetime = __object
        elif isinstance(__object, dt.date): __datetime = dt.datetime(*__object.timetuple()[:6])
        elif get_ts_unit(__object): return from_timestamp(__object, default, tzinfo, astimezone)
        else: __datetime = dateparse(__object, yearfirst=True)
        return set_timezone(__datetime, tzinfo, astimezone, droptz)
    except CastError: return default


def cast_time(__object: DateFormat, default=None, tzinfo: Optional[Timezone]=None,
                astimezone: Optional[Timezone]=None) -> dt.time:
    __datetime = cast_datetime(__object, default, tzinfo, astimezone, droptz=False)
    if isinstance(__datetime, dt.datetime): return __datetime.time()


def cast_timestamp(__object: DateFormat, default=None, tzinfo: Optional[Timezone]=None,
                    astimezone: Optional[Timezone]=None, tsUnit: Literal["ms","s"]="ms") -> int:
    __datetime = cast_datetime(__object, default, tzinfo, astimezone, droptz=False)
    if isinstance(__datetime, dt.datetime):
        return int(__datetime.timestamp()*(1000 if tsUnit == "ms" else 1))


def cast_date(__object: DateFormat, default=None, from_ordinal=False) -> dt.date:
    try:
        if from_ordinal: return dt.date.fromordinal(cast_int(__object, clean=False))
        elif isinstance(__object, dt.datetime): return __object.date()
        elif isinstance(__object, dt.date): return __object
        else: return dateparse(__object, yearfirst=True).date()
    except CastError: return default


def to_datetime(__object: DateFormat, __type: TypeHint="auto", default=None,
                tzinfo: Optional[Timezone]=None, astimezone: Optional[Timezone]=None,
                droptz=False, tsUnit: Literal["ms","s"]="ms", from_ordinal=False, **kwargs) -> DateNumeric:
    if __type == "auto": pass
    elif is_datetime_type(__type): return cast_datetime(__object, tzinfo=tzinfo, astimezone=astimezone, droptz=droptz)
    elif is_time_type(__type): return cast_time(__object, tzinfo=tzinfo, astimezone=astimezone)
    elif is_timestamp_type(__type): return cast_timestamp(__object, tzinfo=tzinfo, astimezone=astimezone, tsUnit=tsUnit)
    elif is_date_type(__type): return cast_date(__object, from_ordinal=from_ordinal)
    else: return default
    __datetime = cast_datetime(__object, tzinfo=tzinfo, astimezone=astimezone)
    if isinstance(__datetime, dt.datetime):
        return __datetime.date() if __datetime.time() == dt.time(0,0) else __datetime


###################################################################
########################## Datetime Forat #########################
###################################################################

COMMON_DATE_PATTERN = r"\d{4}-\d{2}-\d{2}"
COMMON_DATETIME_PATTERN = r"^(?:\d{4}-\d{2}-\d{2}(?: \d{2}:\d{2}:\d{2}(?:\.\d{1,6})?(?:[+-]\d{4})?)?|\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(?:\.\d{1,6})?(?:[+-]\d{4})?)$"


def is_datetime_format(__object: DateFormat, parse=False) -> bool:
    if isinstance(__object, str) and re.fullmatch(COMMON_DATETIME_PATTERN, __object):
        try: return bool(dateparse(__object)) if parse else True
        except: return False
    else: return False


def get_datetime_format(__object: DateFormat, parse=False) -> str:
    if is_datetime_format(__object, parse=parse):
        return "date" if re.fullmatch(COMMON_DATE_PATTERN, __object) else "datetime"


def cast_datetime_format(__object: DateFormat, default=None, parse=False) -> DateNumeric:
    format = get_datetime_format(__object, parse=parse)
    if format == "date": return cast_date(__object, default=default)
    elif format == "datetime": return cast_datetime(__object, default=default)
    else: return default


###################################################################
############################ Array Type ###########################
###################################################################

def cast_str(__object, default=str(), drop_empty=False, strip=False,
            match: Optional[Union[RegexFormat,MatchFunction]]=None) -> str:
    if match:
        if isinstance(match, Callable) and match(__object): pass
        elif isinstance(match, re.Pattern) and match.search(str(__object)): pass
        elif isinstance(match, str) and re.search(match, str(__object)): pass
        else: return default
    elif (exists(__object) if drop_empty else notna(__object)):
        return str(__object).strip() if strip else str(__object)
    else: return default


def cast_id(__object, default=None) -> str:
    return cast_str(cast_int(__object, default=None), default)


def cast_list(__object, drop_empty=False, iter_type: _TYPE=(List,Set,Tuple)) -> List:
    if isinstance(__object, List): return [__e for __e in __object if exists(__e)] if drop_empty else __object
    elif isinstance(__object, iter_type): return [__e for __e in __object if exists(__e)] if drop_empty else list(__object)
    elif (exists(__object) if drop_empty else notna(__object)): return [__object]
    else: return list()


def cast_tuple(__object, drop_empty=False, iter_type: _TYPE=(List,Set,Tuple)) -> Tuple:
    if isinstance(__object, Tuple): return tuple(__e for __e in __object if exists(__e)) if drop_empty else __object
    elif isinstance(__object, iter_type): return tuple(__e for __e in __object if exists(__e)) if drop_empty else tuple(__object)
    elif (exists(__object) if drop_empty else notna(__object)): return (__object,)
    else: return tuple()


def cast_set(__object, drop_empty=False, iter_type: _TYPE=(List,Set,Tuple)) -> Set:
    if isinstance(__object, Set): return {__e for __e in __object if exists(__e)} if drop_empty else __object
    elif isinstance(__object, iter_type): return {__e for __e in __object if exists(__e)} if drop_empty else set(__object)
    elif (exists(__object) if drop_empty else notna(__object)): return {__object}
    else: return set()


###################################################################
############################ Multitype ############################
###################################################################

def cast_object(__object, __type: TypeHint, default=None, clean=True, drop_empty=False, strip=False,
                match: Optional[Union[RegexFormat,MatchFunction]]=None, trunc: Optional[int]=None,
                tzinfo: Optional[Timezone]=None, astimezone: Optional[Timezone]=None,
                droptz=False, tsUnit: Literal["ms","s"]="ms", from_ordinal=False,
                iter_type: _TYPE=(List,Set,Tuple), **kwargs) -> Any:
    if is_str_type(__type): return cast_str(__object, default=default, drop_empty=drop_empty, strip=strip, match=match)
    elif is_int_type(__type): return cast_int(__object, default=default, clean=clean)
    elif is_float_type(__type): return cast_float(__object, default=default, clean=clean, trunc=trunc)
    elif is_bool_type(__type): return exists(__object) if drop_empty else notna(__object)
    elif is_datetime_type(__type): return cast_datetime(__object, default=default, tzinfo=tzinfo, astimezone=astimezone, droptz=droptz)
    elif is_date_type(__type): return cast_date(__object, default=default, from_ordinal=from_ordinal)
    elif is_list_type(__type): return cast_list(__object, drop_empty=drop_empty, iter_type=iter_type)
    elif is_tuple_type(__type): return cast_tuple(__object, drop_empty=drop_empty, iter_type=iter_type)
    elif is_set_type(__type): return cast_set(__object, drop_empty=drop_empty, iter_type=iter_type)
    elif is_time_type(__type): return cast_time(__object, default=default, tzinfo=tzinfo, astimezone=astimezone)
    elif is_timestamp_type(__type): return cast_timestamp(__object, default=default, tzinfo=tzinfo, astimezone=astimezone, tsUnit=tsUnit)
    elif isinstance(__object, get_type(__type)): return __object
    else: return get_type(__type)(default)



from gscraper.base.types import TypeHint, Logic, Unit, DateFormat, Timedelta, Timezone, CastError
from gscraper.base.types import is_type, is_str_type, is_timestamp_type, INTEGER_TYPES

from gscraper.utils.cast import cast_datetime, cast_date, get_timezone
from gscraper.utils.map import isin, drop_dict, get_scala

from typing import Dict, List, Literal, Optional, Sequence, Tuple, Union
from pandas.tseries.frequencies import to_offset
import datetime as dt
import pandas as pd
import random
import re

from holidays.countries import *
from holidays import country_holidays

from workalendar.core import Calendar
from workalendar.europe import *
from workalendar.usa import *
from workalendar.america import *
from workalendar.africa import *
from workalendar.asia import *
from workalendar.oceania import *
from workalendar.registry import registry


UTC = "UTC"
EST = "US/Eastern"
KST = "Asia/Seoul"

US = "US"
KR = "KR"

DATE_UNIT = ["second", "minute", "hour", "day", "month", "year"]

DATETIME_FORMAT = {
    "date": "%Y-%m-%d",
    "datetime": "%Y-%m-%d %H:%M:%S",
    "datetime_ms": "%Y-%m-%d %H:%M:%S.%f",
    "timezone": "%Y-%m-%d %H:%M:%S%z",
    "timezone_ms": "%Y-%m-%d %H:%M:%S.%f%z",
    "js": "%Y-%m-%dT%H:%M:%S.%f%z",
}

DATETIME_PATTERN = {
    "date": r"^\d{4}-\d{2}-\d{2}$",
    "datetime": r"^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}$",
    "datetime_ms": r"^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d+$",
    "timezone": r"^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\+\d{2}:\d{2}$",
    "timezone_ms": r"^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d+\+\d{2}:\d{2}$",
    "js": r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\+\d{2}:\d{2}$",
}

DATE_RANGE_MSG = "Of the four parameters: start, end, periods, and freq, exactly three must be specified."
INVALID_INTERVAL_MSG = lambda interval: f"'{interval}' is not valid date interval for pandas date range."


###################################################################
######################### Current Datetime ########################
###################################################################

def trunc_datetime(__datetime: dt.datetime, unit: Literal["second","minute","hour","day","month","year"]=str()) -> dt.datetime:
    if unit not in DATE_UNIT: return __datetime
    index = DATE_UNIT.index(unit.lower())
    if index >= 0: __datetime = __datetime.replace(microsecond=0)
    if index >= 1: __datetime = __datetime.replace(second=0)
    if index >= 2: __datetime = __datetime.replace(minute=0)
    if index >= 3: __datetime = __datetime.replace(hour=0)
    if index >= 4: __datetime = __datetime.replace(day=1)
    if index >= 5: __datetime = __datetime.replace(month=1)
    return __datetime


def now(__format=str(), days=0, seconds=0, microseconds=0, milliseconds=0, minutes=0,
        hours=0, weeks=0, tzinfo=None, droptz=False, droptime=False, 
        unit: Literal["second","minute","hour","day","month","year"]="second") -> Union[dt.datetime,dt.date,str]:
    try: delta = dt.timedelta(days, seconds, microseconds, milliseconds, minutes, hours, weeks)
    except CastError: return

    __datetime = dt.datetime.now(get_timezone(tzinfo)) - delta
    if droptz: __datetime = __datetime.replace(tzinfo=None)
    if unit: __datetime = trunc_datetime(__datetime, unit=unit)
    if droptime: __datetime = __datetime.date()
    return __datetime.strftime(__format) if __format else __datetime


def today(__format=str(), days=0, weeks=0, tzinfo=None, droptime=True) -> Union[dt.date,dt.datetime,str]:
    return now(__format, days=days, weeks=weeks, tzinfo=tzinfo, droptime=droptime, unit="day")


###################################################################
########################### Get Datetime ##########################
###################################################################

def get_datetime(__object: Optional[DateFormat]=None, if_null: Optional[Union[int,str]]=0,
                days=0, seconds=0, microseconds=0, milliseconds=0, minutes=0, hours=0, weeks=0,
                tzinfo=None, astimezone=None, droptz=False,
                unit: Literal["second","minute","hour","day","month","year"]="second") -> dt.datetime:
    __datetime = cast_datetime(__object, tzinfo=tzinfo, astimezone=astimezone, droptz=droptz)
    if not isinstance(__datetime, dt.datetime):
        if isinstance(__object, int): __datetime = now(hours=__object, tzinfo=tzinfo, droptz=droptz)
        elif isinstance(if_null, int): __datetime = now(hours=if_null, tzinfo=tzinfo, droptz=droptz)
        elif isinstance(if_null, str): __datetime = cast_datetime(__object, tzinfo=tzinfo, astimezone=astimezone, droptz=droptz)
    if isinstance(__datetime, dt.datetime):
        __datetime = __datetime - dt.timedelta(days, seconds, microseconds, milliseconds, minutes, hours, weeks)
        return trunc_datetime(__datetime, unit=unit) if unit else __datetime


def get_datetime_pair(startTime: Optional[DateFormat]=None, endTime: Optional[DateFormat]=None,
                    if_null: Optional[Unit]=None, tzinfo=None, astimezone=None, droptz=False,
                    unit: Literal["second","minute","hour","day","month","year"]="second") -> Tuple[dt.datetime,dt.datetime]:
    context = dict(tzinfo=tzinfo, astimezone=astimezone, droptz=droptz, unit=unit)
    startTime = get_datetime(startTime, if_null=get_scala(if_null, 0), **context)
    endTime = get_datetime(endTime, if_null=get_scala(if_null, 1), **context)
    __min = min(startTime, endTime) if startTime and endTime else startTime
    __max = max(startTime, endTime) if startTime and endTime else endTime
    return __min, __max


def get_time(__object: Optional[DateFormat]=None, if_null: Optional[Union[int,str]]=0,
            seconds=0, microseconds=0, milliseconds=0, minutes=0, hours=0, tzinfo=None, astimezone=None,
            unit: Literal["second","minute","hour","day","month","year"]="second") -> dt.time:
    __datetime = get_datetime(**locals())
    if isinstance(__datetime, dt.datetime):
        return __datetime.time()


def get_timestamp(__object: Optional[DateFormat]=None, if_null: Optional[Union[int,str]]=0,
                days=0, seconds=0, microseconds=0, milliseconds=0, minutes=0, hours=0, weeks=0,
                tzinfo=None, astimezone=None, droptz=False, tsUnit: Literal["ms","s"]="ms",
                unit: Literal["second","minute","hour","day","month","year"]="second") -> int:
    __datetime = get_datetime(**drop_dict(locals(), "tsUnit", inplace=False))
    if isinstance(__datetime, dt.datetime):
        return int(__datetime.timestamp()*(1000 if tsUnit == "ms" else 1))


def get_random_seconds(start: Union[float,int], stop: Optional[Union[float,int]]=None, step=1) -> Union[float,int]:
    has_millis = isinstance(start, float) or isinstance(stop, float)
    if has_millis:
        start = int(start * 1000)
        stop = int(stop * 1000) if isinstance(stop, (float,int)) else None
    secs = random.randrange(start, stop, step)
    return secs/1000 if has_millis else secs


def get_date(__object: Optional[DateFormat]=None, if_null: Optional[Union[int,str]]=0,
            days=0, weeks=0, tzinfo=None, busdate=False, country_code=str()) -> dt.date:
    __date = cast_date(__object)
    if not isinstance(__date, dt.date):
        if isinstance(__object, int): __date = today(days=__object, tzinfo=tzinfo)
        elif isinstance(if_null, int): __date = today(days=if_null, tzinfo=tzinfo)
        elif isinstance(if_null, str): __date = cast_date(if_null)
    if isinstance(__date, dt.date):
        __date = __date - dt.timedelta(days=days, weeks=weeks)
        return get_last_working_day(__date, country_code) if busdate and country_code else __date


def get_date_pair(startDate: Optional[DateFormat]=None, endDate: Optional[DateFormat]=None,
                if_null: Optional[Unit]=None, tzinfo=None, busdate: Optional[Logic]=False, country_code=str()) -> Tuple[dt.date,dt.date]:
    context = dict(tzinfo=tzinfo, country_code=country_code)
    startDate = get_date(startDate, if_null=get_scala(if_null, 0), busdate=get_scala(busdate, 0), **context)
    endDate = get_date(endDate, if_null=get_scala(if_null, 1), busdate=get_scala(busdate, 1), **context)
    __min = min(startDate, endDate) if startDate and endDate else startDate
    __max = max(startDate, endDate) if startDate and endDate else endDate
    return __min, __max


def set_datetime(__datetime: dt.datetime, __format="%Y-%m-%d %H:%M:%S",
                __type: TypeHint=str, tsUnit: Literal["ms","s"]="ms") -> Union[str,int]:
    if not isinstance(__datetime, dt.datetime): return str()
    if __format: return __datetime.strftime(DATETIME_FORMAT.get(__format,__format))
    elif is_str_type(__type): return str(__datetime)
    elif is_timestamp_type(__type): return get_timestamp(__datetime, tsUnit=tsUnit)
    else: return


def set_date(__date: dt.date, __format="%Y-%m-%d", __type: TypeHint=str) -> Union[str,int]:
    if not isinstance(__date, dt.date): return str()
    if __format: return __date.strftime(DATETIME_FORMAT.get(__format,__format))
    elif is_str_type(__type): return str(__date)
    elif is_type(__type, INTEGER_TYPES+["ordinal"]): return __date.toordinal()
    else: return

###################################################################
############################# Holidays ############################
###################################################################

def get_calendar(country_code: str) -> Calendar:
    return registry.get_calendars()[country_code]()


def get_holidays(country_code: str) -> Dict:
    return country_holidays(country_code)


def is_holiday(__date: dt.date, country_code: str) -> bool:
    return (__date in get_holidays(country_code)) or (not get_calendar(country_code).is_working_day(__date))


def is_working_day(__date: dt.date, country_code: str) -> bool:
    return (__date not in get_holidays(country_code)) and get_calendar(country_code).is_working_day(__date)


def get_last_working_day(__date: dt.date, country_code: str, how: Literal["backward","forward"]="backward") -> dt.date:
    calendar, holidays = get_calendar(country_code), get_holidays(country_code)
    while (__date in holidays) or (not calendar.is_working_day(__date)):
        __date = __date + (dt.timedelta(days=1) * (-1 if how == "backward" else 1))
    return __date


###################################################################
######################### Pandas Frequency ########################
###################################################################

PANDAS_FREQUENCY = {
    "B": "business day frequency",
    "C": "custom business day frequency",
    "D": "calendar day frequency",
    "W": "weekly frequency",
    "M": "month end frequency",
    "SM": "semi-month end frequency (15th and end of month)",
    "BM": "business month end frequency",
    "CBM": "custom business month end frequency",
    "MS": "month start frequency",
    "SMS": "semi-month start frequency (1st and 15th)",
    "BMS": "business month start frequency",
    "CBMS": "custom business month start frequency",
    "Q": "quarter end frequency",
    "BQ": "business quarter end frequency",
    "QS": "quarter start frequency",
    "BQS": "business quarter start frequency",
    "A": "year end frequency",
    "Y": "year end frequency",
    "BA": "business year end frequency",
    "BY": "business year end frequency",
    "AS": "year start frequency",
    "YS": "year start frequency",
    "BAS": "business year start frequency",
    "BYS": "business year start frequency",
    "BH": "business hour frequency",
    "H": "hourly frequency",
    "T": "minutely frequency",
    "min": "minutely frequency",
    "S": "secondly frequency",
    "L": "milliseconds",
    "ms": "milliseconds",
    "U": "microseconds",
    "us": "microseconds",
    "N": "nanoseconds",
}

PANDAS_ANCHORED_FREQUENCY = {
    "W-SUN": "weekly frequency (Sundays). Same as 'W'",
    "W-MON": "weekly frequency (Mondays)",
    "W-TUE": "weekly frequency (Tuesdays)",
    "W-WED": "weekly frequency (Wednesdays)",
    "W-THU": "weekly frequency (Thursdays)",
    "W-FRI": "weekly frequency (Fridays)",
    "W-SAT": "weekly frequency (Saturdays)",
    "(B)Q(S)-DEC": "quarterly frequency, year ends in December. Same as 'Q'",
    "(B)Q(S)-JAN": "quarterly frequency, year ends in January",
    "(B)Q(S)-FEB": "quarterly frequency, year ends in February",
    "(B)Q(S)-MAR": "quarterly frequency, year ends in March",
    "(B)Q(S)-APR": "quarterly frequency, year ends in April",
    "(B)Q(S)-MAY": "quarterly frequency, year ends in May",
    "(B)Q(S)-JUN": "quarterly frequency, year ends in June",
    "(B)Q(S)-JUL": "quarterly frequency, year ends in July",
    "(B)Q(S)-AUG": "quarterly frequency, year ends in August",
    "(B)Q(S)-SEP": "quarterly frequency, year ends in September",
    "(B)Q(S)-OCT": "quarterly frequency, year ends in October",
    "(B)Q(S)-NOV": "quarterly frequency, year ends in November",
    "(B)A(S)-DEC": "annual frequency, anchored end of December. Same as 'A'",
    "(B)A(S)-JAN": "annual frequency, anchored end of January",
    "(B)A(S)-FEB": "annual frequency, anchored end of February",
    "(B)A(S)-MAR": "annual frequency, anchored end of March",
    "(B)A(S)-APR": "annual frequency, anchored end of April",
    "(B)A(S)-MAY": "annual frequency, anchored end of May",
    "(B)A(S)-JUN": "annual frequency, anchored end of June",
    "(B)A(S)-JUL": "annual frequency, anchored end of July",
    "(B)A(S)-AUG": "annual frequency, anchored end of August",
    "(B)A(S)-SEP": "annual frequency, anchored end of September",
    "(B)A(S)-OCT": "annual frequency, anchored end of October",
    "(B)A(S)-NOV": "annual frequency, anchored end of November",
}

_map_freq_pattern = lambda freq: re.sub(r"\((\w+)\)", r"[\1]{,1}", freq)


def flip_pandas_frequency(freq: str, sep='') -> str:
    if re.search(r"(M|Q|Y|A)S"+sep, freq):
        return freq
    elif re.search(r"(M|Q|Y|A)E"+sep, freq):
        return re.sub(r"(M|Q|Y|A)E"+sep, r'\1'+sep, freq)
    else: return re.sub(r"(M|Q|Y|A)"+sep, r'\1S'+sep, freq)


def map_pandas_frequency(freq: str, repl: Optional[str]=None, upper=False, flip=False) -> str:
    if isinstance(repl, str): freq = re.sub(r"\d+", repl, freq).strip()
    if upper and not isin(freq, include=("min","ms","us"), how="any"): freq = freq.upper()
    if flip: freq = flip_pandas_frequency(freq, sep=('-' if '-' in freq else ''))
    return freq


def _is_pandas_str_frequency(freq: str, upper=False, flip=False) -> bool:
    context = dict(upper=upper, flip=flip)
    return (
        (map_pandas_frequency(freq, repl='', **context) in PANDAS_FREQUENCY) or
        (freq.upper() in PANDAS_ANCHORED_FREQUENCY) or
        all([__i in PANDAS_FREQUENCY for __i in map_pandas_frequency(freq, repl=' ', **context).split(' ')]) or
        (freq.upper() in map(_map_freq_pattern, PANDAS_ANCHORED_FREQUENCY.keys())))


def is_pandas_frequency(freq: Timedelta, upper=False, flip=False) -> bool:
    if isinstance(freq, dt.timedelta): return True
    elif freq and isinstance(freq, str):
        return _is_pandas_str_frequency(freq, upper, flip)
    else: return False


def is_daily_frequency(freq: Timedelta) -> bool:
    if isinstance(freq, dt.timedelta): return (freq.days == 1) and (freq.seconds == 0)
    elif isinstance(freq, str): return to_offset(freq).freqstr == "D"
    else: return False


def _validate_pandas_str_frequency(freq: str, upper=True, flip=True) -> str:
    freq = map_pandas_frequency(freq, upper=upper, flip=flip)
    if is_pandas_frequency(freq): return freq
    else: raise ValueError(INVALID_INTERVAL_MSG(freq))


def validate_pandas_frequency(freq: Timedelta, upper=True, flip=True) -> Timedelta:
    if isinstance(freq, str):
        return _validate_pandas_str_frequency(freq, upper=upper, flip=flip)
    elif isinstance(freq, dt.timedelta): return freq
    elif isinstance(freq, int): return dt.timedelta(days=freq)
    else: raise ValueError(INVALID_INTERVAL_MSG(freq))


###################################################################
######################## Pandas Date Range ########################
###################################################################

def trunc_date(__date: dt.date, interval: Timedelta="D", how: Literal["backward","forward"]="backward") -> dt.date:
    if is_daily_frequency(interval) or not isinstance(__date, dt.date): return __date
    offset = to_offset(interval)
    if how == "backward": return offset.rollback(__date).date()
    delta = offset.rollforward(__date).date()
    return offset.rollforward(__date+dt.timedelta(days=1)).date() if delta == __date else delta


def _pair_date_range(date_range: Sequence[dt.date], interval: Timedelta="D") -> Sequence[Union[dt.date,dt.date]]:
    if is_daily_frequency(interval) or not date_range:
        return [(date, date) for date in date_range]
    return [(start, (end-dt.timedelta(days=1))) for start, end in zip(date_range[:-1],date_range[1:])]


def get_date_range(startDate: Optional[DateFormat]=None, endDate: Optional[DateFormat]=None,
                    periods: Optional[int]=None, interval: Timedelta="D", upper=True, flip=True,
                    tzinfo: Optional[Timezone]=None, paired=False) -> List[Union[dt.date,Tuple[dt.date,dt.date]]]:
    if sum(map((lambda x: x is not None), (startDate, endDate, periods, interval))) < 3:
        raise ValueError(DATE_RANGE_MSG)
    interval = validate_pandas_frequency(interval, upper=upper, flip=flip)
    startDate, endDate = trunc_date(get_date(startDate), interval, how="backward"), get_date(endDate, if_null=None)
    date_range = [date.date() for date in pd.date_range(startDate, endDate, periods, interval, tzinfo)]
    if (not isinstance(endDate, dt.date)) and isinstance(periods, int):
        endDate = pd.date_range(startDate, endDate, periods+1, interval, tzinfo)[-1].date() - dt.timedelta(days=1)
    if date_range and (not is_daily_frequency(interval)):
        date_range.append(endDate + dt.timedelta(days=int(paired)))
    return _pair_date_range(date_range, interval) if paired else date_range



from __future__ import annotations
from gscraper.base.types import _KT, _VT, TypeHint, MatchFunction
from gscraper.base.types import Records, RenameMap, TypeMap, get_type, is_array, allin_instance, is_records

from gscraper.utils import notna
from gscraper.utils.map import hier_get, notna_dict, exists_dict, drop_dict
from gscraper.utils.map import kloc, vloc, match_records, drop_duplicates

from abc import ABCMeta
import copy
import functools

from typing import Any, Callable, Dict, Iterable, List, Literal, Optional, Sequence, Tuple, Union
import pandas as pd
import re


NAME, TYPE, DESC, DEFAULT, ALIAS = "name", "type", "desc", "default", "alias"
__NONE__ = "__NONE__"
__OPTIONAL__ = "__OPTIONAL__"

INVALID_MSG = lambda __name: f"'{__name}' is not valid."
INVALID_OBJECT_MSG = lambda __object, __name: f"'{__object}' is not a valid {__name} object."
INVALID_OBJECT_TYPE_MSG = lambda __object, __type: f"'{type(__object)}' is not a valid type for {__type} object."
INVALID_INSTANCE_MSG = lambda __object, __instance: f"'{type(__object)}' is not a valid instance for {__instance} type."


###################################################################
############################# Context #############################
###################################################################

BASE_CONTEXT = lambda self=None, operation=None, info=None, initTime=None, prefix=None, rename=None, \
                        inplace=None, self_var=None, update=None, **context: context


LOG_CONTEXT = lambda logger=None, logJson=None, interruptType=None, errorType=None, errors=None, func=None, **context: context


ITERATOR_CONTEXT = lambda iterator=None, iterateArgs=None, iterateCount=None, iterateProduct=None, pagination=None, \
                        pageFrom=None, offsetFrom=None, pageUnit=None, pageLimit=None, fromNow=None, __i=None, \
                        **context: context


MAP_CONTEXT = lambda flow=None, schema=None, responseType=None, match=None, root=None, groupby=None, \
                    countby=None, __index=None, **context: context


SPIDER_CONTEXT = lambda asyncio=None, host=None, field=None, ssl=None, mappedReturn=None, \
                        maxLimit=None, redirectLimit=None, **context: context


ENCRYPTED_CONTEXT = lambda decryptedKey=None, auth=None, authKey=None, includeCookies=None, **context: context


PIPELINE_CONTEXT = lambda  derivFields=None, globalMessage=None, globalProgress=None, taskProgress=None, \
                            asyncMessage=None, asyncProgress=None, taskErrors=None, dags=None, **context: context


UNIQUE_CONTEXT = lambda **context: \
    PIPELINE_CONTEXT(**ENCRYPTED_CONTEXT(**SPIDER_CONTEXT(**MAP_CONTEXT(**ITERATOR_CONTEXT(**LOG_CONTEXT(**BASE_CONTEXT(**context)))))))


PARAMS_CONTEXT = lambda init=None, data=None, task=None, worker=None, locals=None, which=None, where=None, by=None, verb=None, \
                        default=None, clean=None, how=None, alignment=None, dropna=None, drop_empty=None, unique=None, drop=None, \
                        index=None, log=None, depth=None, hier=None, to=None, countPath=None, hasSize=None, primary=None, **context: context


REQUEST_CONTEXT = lambda session=None, semaphore=None, method=None, url=None, referer=None, messages=None, \
                        params=None, encode=None, data=None, json=None, headers=None, cookies=None, \
                        allow_redirects=None, validate=None, exception=None, valid=None, invalid=None, \
                        close=None, encoding=None, features=None, table_type=None, table_idx=None, table_header=None, \
                        engine=None, **context: context


RESPONSE_CONTEXT = lambda response=None, tzinfo=None, countryCode=None, iterateUnit=None, logName=None, logLevel=None, logFile=None, \
                        delay=None, progress=None, message=None, numTasks=None, apiRedirect=None, redirectUnit=None, **context: context


GCLOUD_CONTEXT = lambda name=None, data=None, account=None, key=None, sheet=None, table=None, query=None, project_id=None, \
                        columns=None, mode=None, primary_key=None, base=None, read=None, upload=None, \
                        cell=None, clear=None, default=None, head=None, headers=None, str_cols=None, arr_cols=None, \
                        rename=None, to=None, progress=None, partition=None, fields=None, **context: context


UPLOAD_CONTEXT = lambda queryList=None, uploadList=None, **context: context


TASK_CONTEXT = lambda **context: UPLOAD_CONTEXT(**PARAMS_CONTEXT(**context))


SESSION_CONTEXT = lambda session=None, semaphore=None, cookies=str(), **context: \
                        dict(UPLOAD_CONTEXT(**REQUEST_CONTEXT(**PARAMS_CONTEXT(**context))), **exists_dict(cookies=cookies))


PROXY_CONTEXT = lambda session=None, semaphore=None, **context: UNIQUE_CONTEXT(**UPLOAD_CONTEXT(**context))


LOCAL_CONTEXT = lambda __index=None, apiRedirect=None, returnType=None, localSave=None, **context: SESSION_CONTEXT(**context)


REDIRECT_CONTEXT = lambda logFile=None, **context: LOCAL_CONTEXT(**SESSION_CONTEXT(**context))


###################################################################
########################### Custom Dict ###########################
###################################################################

class CustomDict(dict):
    __metaclass__ = ABCMeta

    def __init__(self, __m: Dict=dict(), self_var=True, **kwargs):
        super().__init__()
        self.update(__m, self_var=self_var, **kwargs)

    def copy(self) -> CustomDict:
        return copy.deepcopy(self)

    def update(self, __m: Dict=dict(), inplace=True, self_var=False, **kwargs) -> CustomDict:
        if not inplace: self = self.copy()
        for __key, __value in dict((dict(self.__dict__, **__m) if self_var else __m), **kwargs).items():
            self[__key] = __value
        if not inplace: return self

    def update_notna(self, __m: Dict=dict(), null_if: Dict=dict(), inplace=True, self_var=False, **kwargs) -> CustomDict:
        return self.update(notna_dict(dict(__m, **kwargs), null_if=null_if), inplace=inplace, self_var=self_var)

    def update_exists(self, __m: Dict=dict(), null_if: Dict=dict(), inplace=True, self_var=False, **kwargs) -> CustomDict:
        return self.update(exists_dict(dict(__m, **kwargs), null_if=null_if), inplace=inplace, self_var=self_var)

    def get(self, __key: _KT, default=None, if_null: Literal["drop","pass"]="pass", values_only=True) -> Union[_VT,Dict]:
        if not is_array(__key): return self[__key] if __key in self else default
        elif if_null == "drop": __m = {__k: self[__k] for __k in __key if __k in self}
        else: __m = {__k: (self[__k] if __k in self else default) for __k in __key}
        return list(__m.values()) if values_only else __m

    def print(self, *__object, path: Optional[_KT]=None, drop: Optional[_KT]=None, indent=2, step=2, double_quotes=False, sep=' '):
        __object = __object if __object else (self,)
        pretty_print(*__object, path=path, drop=drop, indent=indent, step=step, double_quotes=double_quotes, sep=sep)

    def __setitem__(self, __key: _KT, __value: _VT):
        super().__setitem__(__key, __value)
        setattr(self, __key, __value)


class OptionalDict(CustomDict):
    __metaclass__ = ABCMeta

    def __init__(self, optional: Dict=dict(), null_if: Dict=dict(), self_var=False, **kwargs):
        super().__init__(kwargs, **notna_dict(optional, null_if=null_if), self_var=self_var)


class TypedDict(CustomDict):
    __metaclass__ = ABCMeta
    dtype = None
    typeCheck = True

    def __init__(self, **kwargs):
        if not self.typeCheck: dict.__init__(self, kwargs)
        else: dict.__init__(self, {__key: self.validate_dtype(__value) for __key, __value in kwargs.items()})

    def validate_dtype(self, __object) -> Dict:
        if (not self.dtype) or isinstance(__object, self.dtype): return __object
        elif isinstance(__object, self.dtype): return __object
        else: self.raise_dtype_error(__object)

    def raise_dtype_error(self, __object, __type=str()):
        dtype = __type if __type else self.dtype.__name__
        raise TypeError(INVALID_OBJECT_TYPE_MSG(__object, dtype))


###################################################################
########################## Custom Records #########################
###################################################################

class CustomRecords(list):
    __metaclass__ = ABCMeta

    def __init__(self, __iterable: Records):
        super().__init__(__iterable)

    def copy(self) -> CustomRecords:
        return copy.deepcopy(self)

    def update(self, __iterable: Iterable, inplace=True) -> CustomRecords:
        if not inplace: self = self.copy()
        self.clear()
        self.add(__iterable)
        if not inplace: return self

    def add(self, __iterable: Iterable):
        for __i in __iterable:
            self.append(__i)

    def get(self, __key: _KT, default=None, if_null: Literal["drop","pass"]="pass", reorder=True,
            values_only=True, hier=False, key_alias: Sequence[_KT]=list(), axis=0) -> Union[Records,List]:
        context = dict(if_null=if_null, reorder=reorder, values_only=values_only, hier=hier, key_alias=key_alias, axis=axis)
        return vloc(self, __key, default=default, **context)

    def map(self, key: str, value: _KT, default=None, if_null: Literal["drop","pass"]="drop", reorder=True,
            values_only=False, hier=False, key_alias: Sequence[_KT]=list()) -> Dict[_VT,Union[_VT,Dict]]:
        if is_array(value):
            context = dict(if_null=if_null, reorder=reorder, values_only=values_only, hier=hier, key_alias=key_alias)
            return {__m[key]: kloc(__m, value, default=default, **context) for __m in self if key in __m}
        elif if_null == "drop":
            return {__m[key]: __m[value] for __m in self if (key in __m) and (value in __m)}
        else: return {__m[key]: (__m[value] if value in __m else default) for __m in self if key in __m}

    def filter(self, __match: Optional[MatchFunction]=None, inplace=False, **match_by_key) -> CustomRecords:
        if isinstance(__match, Callable):
            return self.update(match_records(self, all_keys=True, match=__match), inplace=inplace)
        elif match_by_key:
            return self.update(match_records(self, **match_by_key), inplace=inplace)
        else: return None if inplace else self

    def unique(self, keys: Optional[_KT]=list(), keep: Literal["fist","last",True,False]="first",
                inplace=False) -> CustomRecords:
        if keep == True: return None if inplace else self
        else: return self.update(drop_duplicates(self, keys, keep=keep), inplace=inplace)

    def print(self, *__object, path: Optional[_KT]=None, drop: Optional[_KT]=None, indent=2, step=2, double_quotes=False, sep=' '):
        __object = __object if __object else (self,)
        pretty_print(*__object, path=path, drop=drop, indent=indent, step=step, double_quotes=double_quotes, sep=sep)


class TypedRecords(CustomRecords):
    __metaclass__ = ABCMeta
    dtype = dict
    typeCheck = True

    def __init__(self, *args: Dict):
        if not self.typeCheck: list.__init__(self, args)
        else: list.__init__(self, [self.validate_dtype(__i) for __i in args])

    def validate_dtype(self, __object) -> Dict:
        if (not self.dtype) or isinstance(__object, self.dtype): return __object
        elif isinstance(__object, Dict): return self.dtype(**__object)
        else: raise self.raise_dtype_error(__object)

    def raise_dtype_error(self, __object, __type=str()):
        dtype = __type if __type else self.dtype.__name__
        raise TypeError(INVALID_OBJECT_TYPE_MSG(__object, dtype))

    def append(self, __object: Dict):
        super().append(self.validate_dtype(__object))


class NamedRecords(TypedRecords):
    __metaclass__ = ABCMeta

    def rename(self, __s: str, to: Optional[Literal["name","desc"]]="desc",
                if_null: Union[Literal["pass"],Any]="pass") -> str:
        renameMap = self.get_rename_map(to=to)
        if renameMap and (__s in renameMap): return renameMap[__s]
        else: return __s if if_null == "pass" else if_null

    def get_rename_map(self, to: Literal["name","desc"]="desc") -> RenameMap:
        key, value = (DESC, NAME) if to == NAME else (NAME, DESC)
        return self.map(key, value)

    def get_type_map(self, key: Literal["name","desc"]="name") -> TypeMap:
        return self.map(key, TYPE)


###################################################################
############################## Value ##############################
###################################################################

class Value(OptionalDict):
    __metaclass__ = ABCMeta
    typeCast = True

    def __init__(self, name: _KT, type: TypeHint, optional: Dict=dict(), null_if: Dict=dict(), self_var=False, **kwargs):
        type = get_type(type) if self.typeCast else type
        super().__init__(name=name, type=type, optional=optional, null_if=null_if, self_var=self_var, **kwargs)


class ValueSet(NamedRecords):
    __metaclass__ = ABCMeta
    dtype = Value
    typeCheck = True


###################################################################
############################# Prettier ############################
###################################################################

def to_default(__object) -> Any:
    if isinstance(__object, Dict):
        return {__key: to_default(__value) for __key, __value in __object.items()}
    elif isinstance(__object, List):
        return [to_default(__e) for __e in __object]
    else: return __object


def _format_quote(func) -> str:
    @functools.wraps(func)
    def wrapper(*args, double_quotes=False, **context):
        __s = func(*args, **context)
        return __s.replace('\'', '\"') if double_quotes and isinstance(__s, str) else __s
    return wrapper


@_format_quote
def pretty_str(__object, indent=2, step=2, double_quotes=False) -> str:
    indent = max(indent, step)
    if isinstance(__object, Value):
        return str({__k: to_default(__v) for __k, __v in __object.items()})
    elif isinstance(__object, CustomDict):
        return '{\n'+',\n'.join([' '*indent+f"'{__k}': {pretty_str(__v, indent=indent+step, step=step)}"
                    for __k, __v in __object.items()])+'\n'+' '*(indent-step)+'}'
    elif isinstance(__object, CustomRecords):
        return '[\n'+',\n'.join([' '*indent+pretty_str(__e, indent=indent+step, step=step)
                for __e in __object])+'\n'+' '*(indent-step)+']'
    elif is_array(__object) and allin_instance(__object, (CustomDict,CustomRecords)):
        __s = str(__object)
        return f'{__s[0]}\n'+',\n'.join([' '*indent+pretty_str(__e, indent=indent+step, step=step)
                for __e in __object])+'\n'+' '*(indent-step)+f'{__s[-1]}'
    elif isinstance(__object, str): return f"'{__object}'"
    else: return str(__object).replace('\n', '\\n')


@_format_quote
def pretty_object(__object, path: Optional[_KT]=None, drop: Optional[_KT]=None, indent=2, step=2, double_quotes=False) -> str:
    if notna(path): __object = hier_get(__object, path)
    if notna(drop): __object = drop_dict(__object, drop)
    context = dict(indent=indent, step=step)
    if isinstance(__object, (CustomDict,CustomRecords)):
        return pretty_str(__object, **context)
    elif isinstance(__object, Dict):
        return pretty_str(CustomDict(__object), **context)
    elif is_records(__object, how="all"):
        return pretty_str(CustomRecords(__object), **context)
    elif isinstance(__object, pd.DataFrame):
        return "pd.DataFrame("+pretty_str(CustomRecords(__object.to_dict("records")), **context)+")"
    elif isinstance(__object, pd.Series):
        return "pd.Series("+pretty_str(__object.tolist(), **context)+")"
    else: return pretty_str(__object)


def pretty_print(*args, path: Optional[_KT]=None, drop: Optional[_KT]=None, indent=2, step=2, double_quotes=False, sep=' '):
    print(sep.join([
        pretty_object(__object, path=path, drop=drop, indent=indent, step=step, double_quotes=double_quotes) for __object in args]))


###################################################################
############################## Query ##############################
###################################################################

class Variable(Value):
    def __init__(self, name: _KT, type: TypeHint, desc: Optional[str]=None, default: Optional[Any]=__NONE__,
                iterable=False, arr_options: Optional[Dict]=None, enum: Union[Tuple[str,...],RenameMap]=None):
        type, iterable = self.validate_iterable(type, iterable)
        super().__init__(name=name, type=type)
        self.update_optional(desc, default, iterable, arr_options, enum)

    def validate_iterable(self, type: TypeHint, iterable=False) -> Union[TypeHint,bool]:
        if isinstance(type, str) and bool(re.match(r"\[[^]]*\]", type)):
            type = re.sub(r"^\[([^]]*)\]$", "\g<1>", type)
            iterable = True
        return type, iterable

    def update_optional(self, desc: Optional[str]=None, default: Optional[Any]=__NONE__, iterable=False,
                        arr_options: Optional[Dict]=None, enum: Union[Tuple[str,...],RenameMap]=None):
        enum = enum if isinstance(enum, (Dict,Tuple)) else None
        if isinstance(arr_options, Dict):
            iterable = True
            if (arr_options.get("default") is not None) and (not arr_options.get("drop_empty")):
                arr_options["dropna"] = True
        elif iterable:
            if enum: arr_options = dict(dropna=True, unique=True)
            else: arr_options = dict(drop_empty=True, unique=True)
        self.update_notna(
            dict(desc=desc, default=default, iterable=iterable, arr_options=arr_options, enum=enum),
            null_if=dict(default=__NONE__))

    def copy(self) -> Variable:
        return copy.deepcopy(self)

    def update(self, __m: Dict=dict(), inplace=True, self_var=False, **kwargs) -> Variable:
        return super().update(__m, inplace=inplace, self_var=self_var, **kwargs)


class Query(ValueSet):
    dtype = Variable
    typeCheck = True

    def __init__(self, *variables: Variable):
        super().__init__(*variables)

    def copy(self) -> Query:
        return copy.deepcopy(self)

    def update(self, __iterable: Iterable[Variable], inplace=True) -> Query:
        return super().update(__iterable, inplace=inplace)

    def get_required_query(self) -> Query:
        optional_query = self.map(key="name", value="default")
        return Query(*[variable for variable in self if variable[NAME] not in optional_query])

    def get_optional_query(self) -> Query:
        optional_query = self.map(key="name", value="default")
        return Query(*[variable for variable in self if variable[NAME] in optional_query])


###################################################################
########################## Session Query ##########################
###################################################################

FILTER_QUERY = lambda mapped=False: Query(
    Variable(name="fields", type="STRING", iterable=(not mapped), default=(dict() if mapped else list())),
    Variable(name="ranges", type="[STRING]", default=list()),
    Variable(name="returnType", type="STRING", default=None),
)

TIME_QUERY = lambda: Query(
    Variable(name="tzinfo", type="STRING", default=None),
    Variable(name="countryCode", type="STRING", default=None),
    Variable(name="datetimeUnit", type="STRING", default=None),
)

LOG_QUERY = lambda: Query(
    Variable(name="logName", type="STRING", default=None),
    Variable(name="logLevel", type="STRING", default="WARN"),
    Variable(name="logFile", type="STRING", default=None),
    Variable(name="localSave", type="BOOLEAN", default=None),
    Variable(name="debug", type="[STRING]", default=None),
    Variable(name="extraSave", type="[STRING]", default=None),
    Variable(name="interrupt", type="[STRING]", default=None),
)

REQUEST_QUERY = lambda: Query(
    Variable(name="numRetries", type="INTEGER", default=0),
    Variable(name="delay", type="[FLOAT]", default=1.),
    Variable(name="cookies", type="STRING", default=None),
)

GCLOUD_QUERY = lambda: Query(
    Variable(name="queryList", type="[DICT]", default=list()),
    Variable(name="uploadList", type="[DICT]", default=list()),
    Variable(name="account", type="STRING", default=None),
)


###################################################################
########################### Spider Query ##########################
###################################################################

ITERATOR_QUERY = lambda: Query(
    Variable(name="iterateUnit", type="INTEGER", default=None),
    Variable(name="interval", type="STRING", default=None),
)

TASK_QUERY = lambda: Query(
    Variable(name="fromNow", type="[INTEGER]", default=None),
    Variable(name="discard", type="BOOLEAN", default=True),
    Variable(name="progress", type="BOOLEAN", default=True),
)

GATHER_QUERY = lambda: Query(
    Variable(name="where", type="STRING", default=str()),
    Variable(name="which", type="STRING", default=str()),
    Variable(name="verb", type="STRING", default=str()),
    Variable(name="by", type="STRING", default=str()),
    Variable(name="message", type="STRING", default=str()),
)


###################################################################
############################ Base Query ###########################
###################################################################

ASYNC_QUERY = lambda: Query(
    Variable(name="numTasks", type="INTEGER", default=100),
    Variable(name="apiRedirect", type="BOOLEAN", default=False),
    Variable(name="redirectUnit", type="INTEGER", default=None),
)

ENCRYPTED_QUERY = lambda: Query(
    Variable(name="encryptedKey", type="STRING", default=None),
    Variable(name="decryptedKey", type="STRING", default=None),
)

PIPELINE_QUERY = lambda: Query(
    Variable(name="globalMessage", type="STRING", default=str()),
    Variable(name="globalProgress", type="BOOLEAN", default=False),
    Variable(name="asyncMessage", type="STRING", default=str()),
    Variable(name="asyncProgress", type="BOOLEAN", default=False),
    Variable(name="taskProgress", type="BOOLEAN", default=True),
)


def get_base_query(asyncio=False, encrypted=False, mapped=False, pipeline=False, **kwargs) -> Query:
    return Query(
        *FILTER_QUERY(mapped),
        *TIME_QUERY(),
        *LOG_QUERY(),
        *REQUEST_QUERY(),
        *(list() if pipeline else TASK_QUERY()),
        *(list() if pipeline else GATHER_QUERY()),
        *(ASYNC_QUERY() if asyncio else list()),
        *GCLOUD_QUERY(),
        *(ENCRYPTED_QUERY() if encrypted else list()),
        *(PIPELINE_QUERY() if pipeline else list()),
    )


###################################################################
########################### Extra Query ###########################
###################################################################

class OptionalQuery(Query):
    dtype = Variable
    typeCheck = True

    def __init__(self, *variables: Variable):
        super().__init__(*[
            variable for variable in variables if isinstance(variable, Dict) and (variable.get(DEFAULT) != __OPTIONAL__)])


DATE_RANGE_QUERY = lambda startDate=__NONE__, endDate=__NONE__, interval=__OPTIONAL__: OptionalQuery(
    Variable(name="startDate", type="DATE", default=startDate),
    Variable(name="endDate", type="DATE", default=endDate),
    Variable(name="interval", type="STRING", default=interval),
)

PAGE_RANGE_QUERY = lambda size=__OPTIONAL__, pageSize=__OPTIONAL__, pageStart=__OPTIONAL__, offset=__OPTIONAL__: OptionalQuery(
    Variable(name="size", type="INTEGER", default=size),
    Variable(name="pageSize", type="INTEGER", default=pageSize),
    Variable(name="pageStart", type="INTEGER", default=pageStart),
    Variable(name="offset", type="INTEGER", default=offset),
)

ISIN_KEYWORD_QUERY = lambda exact=__OPTIONAL__, include=__OPTIONAL__, exclude=__OPTIONAL__: OptionalQuery(
    Variable(name="exact", type="[STRING]", default=exact),
    Variable(name="include", type="[STRING]", default=include),
    Variable(name="exclude", type="[STRING]", default=exclude),
)

LOGIN_QUERY = lambda userid=__NONE__, passwd=__NONE__: Query(
    Variable(name="userid", type="STRING", default=userid),
    Variable(name="passwd", type="STRING", default=passwd),
)
